from decimal import Decimal


def safe_decimal(number) -> Decimal:
    """
    Try to convert any number to Decimal.
    Suited mostly for floats:
    >>> Decimal(1.01)
    Decimal('1.0100000000000000088817841970012523233890533447265625')
    >>> Decimal('1.01')
    Decimal('1.01')
    """
    if isinstance(number, Decimal):
        return number
    elif isinstance(number, float):
        return Decimal(str(number))
    elif isinstance(number, (int, str)):
        return Decimal(number)
    else:
        raise TypeError(f'Type {type(number)} is not supported.')



from blockapi.services import BlockchainAPI


class TzStatsAPI(BlockchainAPI):
    """
    Tezos
    Explorer: https://api.tzstats.com/explorer/
    Docs: https://tzstats.com/docs/api/index.html
    """

    symbol = 'XTZ'
    base_url = 'https://api.tzstats.com/'
    rate_limit = 0
    coef = 1
    max_items_per_page = 50
    page_offset_step = 1

    supported_requests = {
        'get_account': 'explorer/account/{address}'
    }

    def get_balance(self):
        balance = self.request('get_account', address=self.address)
        return [{
            'symbol': self.symbol,
            'amount': float(balance['spendable_balance'])
        }]

    def get_account(self):
        response = self.request('get_account', address=self.address)

        response['manager'] = self.request('get_account',
                                           address=response['manager'])
        response['delegate'] = self.request('get_account',
                                            address=response['delegate'])
        return response



from blockapi.services import BlockchainAPI


class ZchainAPI(BlockchainAPI):
    """
    coins: zcash
    API docs: https://explorer.zcha.in/api
    Explorer: https://explorer.zcha.in/
    """

    active = True

    symbol = 'ZEC'
    base_url = 'https://api.zcha.in'
    rate_limit = 0
    coef = 1
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/v2/mainnet/accounts/{address}',
    }

    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        retval = response.get('balance') * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]



from pytest import mark

from blockapi.api import BlockscoutEthereumClassicAPI, BlockscoutXdaiAPI
from blockapi.test_data import test_addresses


class TestBlockscoutXdaiAPI:
    ADDRESS = test_addresses['STAKE'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = BlockscoutXdaiAPI(self.ADDRESS)
        result = api.get_balance()

        assert len(result) == 4


class TestBlockscoutEthereumClassicAPI:
    ADDRESS = test_addresses['ETC'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = BlockscoutEthereumClassicAPI(self.ADDRESS)
        result = api.get_balance()

        assert len(result) == 3



from decimal import Decimal

from blockapi import (check_address_valid, get_all_supported_coins,
                      get_api_classes_for_coin)
from blockapi.test_data import get_test_api_key, test_addresses


class TestBlockApiProviders:

    def __init__(self, data=None, methods=None):
        """State what currencies should be tested on which addresses
        with which expected values

        Args:
            data (list): list of dicts with structure
            {
                "currency": (str),  # Should be valid symbol
                "address": (str),
                "value": (Decimal)
            }
            methods (list): what methods should be tested, if None
                get_balance is tested
        """

        self.currencies = get_all_supported_coins()
        self.data = data if data is not None else []
        self.methods = methods if methods else ['get_balance']
        self.results = {m: [] for m in self.methods}
        self.diagnostic = None
        self.error_api = []

    def run_all(self):
        """Auto run of all APIs using test data"""
        data = []

        # Create test sample
        for currency in self.currencies:

            if not test_addresses.get(currency):
                continue

            data.append({
                'currency': currency,
                'address': test_addresses[currency][0],
                'value': '0'
            })

        self.data = data
        self.run()

    def run(self):
        """Runs tests for given data"""

        for currency in self.data:
            classes = get_api_classes_for_coin(symbol=currency['currency'])

            for class_ in classes:

                # try to load api_key
                api_key = get_test_api_key(class_.__name__)
                try:
                    class_instance = class_(
                        address=currency['address'],
                        api_key=api_key
                    )
                except Exception as e:
                    self.error_api.append({
                        'api': class_.__name__,
                        'error': 'Init error',
                        'message': e
                    })
                    continue

                for method in self.methods:
                    try:
                        self.results[method].append({
                            'result': getattr(class_instance, method)(),
                            'api': class_.__name__,
                            'currency': currency['currency'],
                            'success': True
                        })
                    except Exception as e:
                        self.error_api.append({
                            'api': class_.__name__,
                            'error': f'{method} error',
                            'message': e
                        })

    def get_diagnostic(self, method=None):
        """Returns info about results"""

        if method is None:
            method = 'get_balance'

        if not self.results:
            return 'Run "run" or "run_all" first.'

        if method not in self.methods:
            return 'Choose valid method.'

        return {
            'api_with_error': [a['api'] for a in self.error_api],
            'api_with_result_none': [r['api'] for r in self.results[method]
                                     if r['result'] is None],
            'api_with_structure_error': [r['api'] for r in self.results[method]
                                         if not isinstance(r['result'], list)],
            'api_with_results': {
                c: {
                    'results': [
                        r for r in self.results[method]
                        if r['result'] is not None
                        and r['currency'] == c
                        and isinstance(r['result'], list)
                    ],
                    'mean': self._mean(method=method, currency=c)
                } for c in self.currencies
            }
        }

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, _data):
        valid_data = []
        not_valid_data = []
        for d in _data:

            if d['currency'] not in self.currencies:
                not_valid_data.append({
                    'message': 'Symbol not valid',
                    'data': d
                })
                continue

            if not check_address_valid(d['currency'], d['address']):
                not_valid_data.append({
                    'message': 'Check address not successful',
                    'data': d
                })
                continue

            try:
                d['value'] = Decimal(d['value'])
            except ValueError:
                not_valid_data.append({
                    'message': 'Value is not a number',
                    'data': d
                })
                continue

            valid_data.append(d)

        self._data = valid_data

    def _mean(self, method, currency):
        records = [r['result'] for r in self.results[method]
                   if r['result'] is not None and r['currency'] == currency
                   and isinstance(r['result'], list)]

        sum_ = sum([
            next((i['amount'] for i in r if i['symbol'] == currency), 0)
            for r in records
        ])
        return sum_ / len(records) if records else 0



import json

from blockapi.services import (
    APIError, BlockchainAPI,
)


class SolanaApi(BlockchainAPI):
    """
    Solana RPC
    API docs: https://docs.solana.com/apps/jsonrpc-api
    """

    symbol = 'SOL'
    base_url = 'https://api.mainnet-beta.solana.com/'
    rate_limit = None
    coef = 1e-9
    start_offset = 0
    max_items_per_page = 1000
    page_offset_step = 1

    # follow used pattern even though this API uses POST requests
    supported_requests = {
        'get_balance': '',
        'get_txs_signatures': '',
        'get_tx': ''
    }

    token_program_id = 'TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA'

    # mainnet tokens
    # https://github.com/solana-labs/explorer/blob/master/src/tokenRegistry.ts
    tokens = {
        'SRMuApVNdxXokk5GT7XD5cUUgXMBCoAz2LHeuAoKWRt': {
            'name': 'Serum',
            'symbol': 'SRM',
        },
        'MSRMcoVyrFxnSgo5uXwone5SKcGhT1KEJMFEkMEWf9L': {
            'name': 'MegaSerum',
            'symbol': 'MSRM',
        },
        '9n4nbM75f5Ui33ZbPYXn59EwSgE8CGsHtAeTH5YFeJ9E': {
            'symbol': 'BTC',
            'name': 'Wrapped Bitcoin',
        },
        '2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk': {
            'symbol': 'ETH',
            'name': 'Wrapped Ethereum',
        },
        'AGFEad2et2ZJif9jaGpdMixQqvW5i81aBdvKe7PHNfz3': {
            'symbol': 'FTT',
            'name': 'Wrapped FTT',
        },
        '3JSf5tPeuscJGtaCp5giEiDhv51gQ4v3zWg8DGgyLfAB': {
            'symbol': 'YFI',
            'name': 'Wrapped YFI',
        },
        'CWE8jPTUYhdCTZYWPTe1o5DFqfdjzWKc9WKz6rSjQUdG': {
            'symbol': 'LINK',
            'name': 'Wrapped Chainlink',
        },
        'Ga2AXHpfAF6mv2ekZwcsJFqu7wB4NV331qNH7fW9Nst8': {
            'symbol': 'XRP',
            'name': 'Wrapped XRP',
        },
        'BQcdHdAQW1hczDbBi9hiegXAR7A98Q9jx3X3iBBBDiq4': {
            'symbol': 'USDT',
            'name': 'Wrapped USDT',
        },
        'BXXkv6z8ykpG1yuvUDPgh732wzVHB69RnB9YgSYh3itW': {
            'symbol': 'USDC',
            'name': 'Wrapped USDC',
        },
        'So11111111111111111111111111111111111111112': {
            'symbol': 'SOL',
            'name': 'Wrapped SOL',
        },
        'AR1Mtgh7zAtxuxGd2XPovXPVjcSdY3i4rQYisNadjfKy': {
            'symbol': 'SUSHI',
            'name': 'Wrapped Sushi',
        },
        'SF3oTvfWzEP3DTwGSvUXRrGTvr75pdZNnBLAH9bzMuX': {
            'symbol': 'SXP',
            'name': 'Wrapped Swipe',
        },
        'CsZ5LZkDS7h9TDKjrbL7VAwQZ9nsRu8vJLhRYfmGaN8K': {
            'symbol': 'ALEPH',
            'name': 'Wrapped Aleph',
        },
    }

    def __init__(self, address, *args, **kwargs):
        super().__init__(address)
        self._headers = {
            'Content-Type': 'application/json'
        }

    def get_balance(self):
        balances = []

        sol_balance = self._get_sol_balance()
        if sol_balance:
            balances.append(sol_balance)

        token_balances = self._get_token_balances()
        if token_balances:
            balances += token_balances

        return balances

    def _get_sol_balance(self):
        response = self._request(
            method='getBalance',
            params=[self.address]
        )

        return {
            'symbol': self.symbol,
            'amount': response['result']['value'] * self.coef
        }

    def _get_token_balances(self):
        response = self._request(
            method='getTokenAccountsByOwner',
            params=[
                self.address,
                {
                    'programId': self.token_program_id
                },
                {
                    'encoding': 'jsonParsed'
                }
            ]
        )

        balances = [self._parse_token_balance(b)
                    for b in response['result']['value']]

        return balances

    def _parse_token_balance(self, raw):
        info = raw['account']['data']['parsed']['info']

        token_address = info['mint']
        token_info = self.tokens.get(token_address)

        _amount = info['tokenAmount']
        if _amount.get('uiAmount'):
            amount = _amount['uiAmount']
        elif _amount.get('decimals'):
            amount = float(_amount['amount']) * pow(10, -_amount['decimals'])
        else:
            amount = _amount['amount']

        return {
            'symbol': token_info['symbol'] if token_info else 'UNKNOWN',
            'address': token_address,
            'amount': amount,
            'name': token_info['name'] if token_info else 'Unknown',
            'account_address': raw['pubkey']
        }

    def get_txs_signatures(self, limit=None):
        if not limit:
            limit = self.max_items_per_page

        response = self._request(
            method='getConfirmedSignaturesForAddress2',
            params=[
                self.address,
                {
                    'limit': limit
                }
            ]
        )

        return [
            {
                'signature': r['signature'],
                'error': bool(r['err'])
            }
            for r in response['result']
        ]

    def get_tx(self, signature):
        response = self._request(
            method='getConfirmedTransaction',
            params=[
                signature,
                'jsonParsed'
            ]
        )
        return response['result']

    def _request(self, method, params):
        body = json.dumps({
            'jsonrpc': '2.0',
            'id': 1,
            'method': method,
            'params': params
        })

        response = self.request(
            # request method is not needed, it's included in body
            request_method=None,
            body=body,
            headers=self._headers
        )
        if response.get('error'):
            raise APIError(response['error'])

        return response



from decimal import Decimal

from pytest import mark

from blockapi.api.subscan import SubscanPolkaAPI, SubscanWestendAPI
from blockapi.test_data import test_addresses


class TestSubscanAPI:
    ADDRESS = test_addresses['WND'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = SubscanWestendAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [{'symbol': 'WND', 'amount': Decimal('1.140399983849')}]

    @mark.vcr()
    def test_get_txs(self):
        api = SubscanWestendAPI(address=self.ADDRESS)
        result = api.get_txs()

        assert isinstance(result, list)
        assert len(result) == 3

    @mark.vcr()
    def test_rewards(self):
        api = SubscanWestendAPI(address=self.ADDRESS)
        result = api.get_rewards()

        assert isinstance(result, list)
        assert result[0]['amount'] == Decimal('4.967762520373')

    @mark.vcr()
    def test_staking(self):
        api = SubscanWestendAPI(address=self.ADDRESS)
        result = api.get_staking()

        assert isinstance(result, Decimal)
        assert result == Decimal('0.98039999247')




from datetime import datetime
from decimal import Decimal

import pytest

from blockapi.api.kyber import KyberAPI

USER_ADDRESS = '0x1375355fEcCB15e0DBedd7dcF3c496E2b1a25f6A'


class TestKyber:

    @staticmethod
    def test_raises_value_error():
        with pytest.raises(ValueError):
            KyberAPI(USER_ADDRESS, network='ropsten_typo')

    @staticmethod
    @pytest.mark.vcr()
    def test_get_staker_epoch_info():
        info = KyberAPI(USER_ADDRESS, network='ropsten') \
            .get_staker_epoch_info(182)
        assert list(info.keys()) == ['stake_amount',
                                     'delegated_stake_amount',
                                     'pending_stake_amount',
                                     'delegate']
        assert [type(val) for val in info.values()] == \
               [Decimal, Decimal, Decimal, str]

    @staticmethod
    @pytest.mark.vcr()
    def test_get_actions():
        actions = KyberAPI(USER_ADDRESS, network='ropsten').get_user_actions()
        assert list(actions[0].keys()) == ['epoch', 'type', 'tx_hash', 'meta',
                                           'action_date']
        assert [type(val) for val in actions[0].values()] == \
               [Decimal, str, str, dict, datetime]

    @staticmethod
    @pytest.mark.vcr()
    def test_get_rewards():
        rewards = KyberAPI(USER_ADDRESS, network='ropsten') \
            .get_staker_rewards()
        assert list(rewards[0].keys()) == ['epoch', 'amount', 'claimed',
                                           'tx_hash', 'total_stake',
                                           'total_reward', 'total_voted']
        assert [type(val) for val in rewards[0].values()] == \
               [Decimal, Decimal, bool, str, Decimal, Decimal, Decimal]

    @staticmethod
    @pytest.mark.vcr()
    def test_get_staker_votes():
        votes = KyberAPI(USER_ADDRESS, network='ropsten').get_staker_votes()
        assert list(votes[0].keys()) == ['staker',
                                         'epoch',
                                         'campaign_id',
                                         'option',
                                         'power']
        assert [type(val) for val in votes[0].values()] == \
               [str, Decimal, Decimal, Decimal, str]



from datetime import datetime
from decimal import Decimal
from typing import Dict, List

from blockapi.services import BlockchainAPI


class KyberAPI(BlockchainAPI):
    """
    Kyber Network
    API docs: https://github.com/KyberNetwork/developer-portal/blob/
                                    stakingSection/api-endpoints.md
    Explorer: https://tracker.kyber.network/#/trades
    """

    symbol = 'KNC'
    supported_requests = {
        'get_staker_epoch_info': 'stakers/{address}?epoch={epoch}',
        'get_user_actions': 'stakers/{address}/actions',
        'get_staker_rewards': 'stakers/{address}/rewards',
        'get_staker_votes': 'stakers/{address}/votes'
    }

    def __init__(self, address: str, network: str = 'mainnet'):
        ENDPOINTS = {
            'mainnet': 'https://api.kyber.org/',
            'ropsten': 'https://ropsten-api.kyber.org/',
        }
        if network not in ENDPOINTS.keys():
            raise ValueError(
                f'Network has to be one of {list(ENDPOINTS.keys())}. '
                f'Received \'{network}\' instead.'
            )
        self.base_url = ENDPOINTS[network]

        super(KyberAPI, self).__init__(address)

    def get_staker_epoch_info(self, epoch: int) -> Dict:
        """
        Get staker information for a given epoch

        return: a dict with the following info:
            {
                "stake_amount": Decimal,
                "delegated_stake_amount": Decimal,
                "pending_stake_amount": Decimal,
                "delegate": str, - Ethereum address of the delegate
            }
        """
        result = self.request('get_staker_epoch_info', address=self.address,
                              epoch=epoch)
        if not result['success'] or not result['data']:
            return {}
        return self._parse_staker_epoch_info(result['data'])

    @staticmethod
    def _parse_staker_epoch_info(info: Dict) -> Dict:
        return {
            'stake_amount': Decimal(info['stake_amount']),
            'delegated_stake_amount': Decimal(info['delegated_stake_amount']),
            'pending_stake_amount': Decimal(info['pending_stake_amount']),
            'delegate': info['delegate']
        }

    def get_user_actions(self) -> List[Dict]:
        """
        Get information about actions undertaken by a given address

        return: a list of dicts with the following info:
            {
                'epoch': Decimal, Epoch number
                'type': str, A type of action executed (e.g. Deposit)
                'tx_hash': str, A tx hash of a given action
                'meta': dict, Additional info (amount, campaign_id, option)
                'action_date': datetime, A date and time of the action
            }
        """
        result = self.request('get_user_actions', address=self.address)
        if not result['success'] or not result['data']:
            return []
        return self._parse_user_actions(result['data'])

    @staticmethod
    def _parse_user_actions(raw_actions: List) -> List[Dict]:
        actions = []
        for action in raw_actions:
            actions.append({
                'epoch': Decimal(action['epoch']),
                'type': action['type'],
                'tx_hash': action['tx_hash'],
                'meta': action['meta'],
                'action_date': datetime.utcfromtimestamp(action['timestamp'])
            })
        return actions

    def get_staker_rewards(self) -> List[Dict]:
        """
        Get information about staking rewards for a given address

        return: a list of dicts with the following info:
            {
                'epoch': Decimal, Epoch number
                'amount': Decimal, ETH reward claimable / claimed for staker
                            and his pool members
                'claimed': bool, Says, whether reward has been claimed
                'tx_hash': str, Transaction hash of the reward claim
                'total_stake': Decimal, Total KNC stake of the staker and other
                                pool members
                'total_reward': Decimal, Total ETH reward claimable by all
                                stakers for epoch
                'total_voted': Decimal, Total vote amount by all stakers for
                                epoch
            }
        """
        result = self.request('get_staker_rewards', address=self.address)
        if not result['success'] or not result['data']:
            return []
        return self._parse_staker_rewards(result['data'])

    @staticmethod
    def _parse_staker_rewards(raw_rewards: List) -> List[Dict]:
        rewards = []
        for reward in raw_rewards:
            rewards.append({
                'epoch': Decimal(reward['epoch']),
                'amount': Decimal(reward['amount']),
                'claimed': reward['claimed'],
                'tx_hash': reward['tx_hash'],
                'total_stake': Decimal(reward['total_stake']),
                'total_reward': Decimal(reward['total_reward']),
                'total_voted': Decimal(reward['total_voted']),
            })
        return rewards

    def get_staker_votes(self) -> List[Dict]:
        """
        Get information about actions undertaken by a given address

        return: a list of dicts with the following info:
            {
                'staker': str, Staker's address
                'epoch': Decimal, Epoch number
                'campaign_id': Decimal, Campaign ID number
                'option': Decimal, Voted option by stake
                'power': str, Voting power of staker (delegations accounted
                for)
            }
        """
        result = self.request('get_staker_votes', address=self.address)
        if not result['success'] or not result['data']:
            return []
        return self._parse_staker_votes(result['data'])

    @staticmethod
    def _parse_staker_votes(raw_votes: List) -> List[Dict]:
        votes = []
        for vote in raw_votes:
            votes.append({
                'staker': vote['staker'],
                'epoch': Decimal(vote['epoch']),
                'campaign_id': Decimal(vote['campaign_id']),
                'option': Decimal(vote['option']),
                'power': vote['power'],
            })
        return votes

    def get_balance(self):
        return None



from pytest import mark

from blockapi.api import SolanaApi
from blockapi.test_data import test_addresses


class TestSolanaAPI:
    ADDRESS = test_addresses['SOL'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = SolanaApi(self.ADDRESS)
        result = api.get_balance()

        assert len(result) == 3
        assert next(b for b in result if b['symbol'] == 'SOL')
        assert next(b for b in result if b['symbol'] == 'YFI')
        usdc_bal = next(b for b in result if b['symbol'] == 'USDT')
        assert usdc_bal['amount'] == 21.17783

    @mark.vcr()
    def test_get_txs_signatures(self):
        api = SolanaApi(self.ADDRESS)
        result = api.get_txs_signatures()

        assert len(result) == 22

    @mark.vcr()
    def test_get_tx(self):
        api = SolanaApi(self.ADDRESS)
        result = api.get_tx(
            '5v1yotDftqHWpgCqDuPoEMaLwyQiHaCiDZ3i2Srf31oKosDRJerk4PQpretG4HkQb'
            'WqRgfSFPLLdKi9NuyGLcqQH'
        )

        assert result



from decimal import Decimal

from blockapi.services import BlockchainAPI


class TerraMoneyApi(BlockchainAPI):
    """
    Terra Money
    API docs: UNKNOWN
    Explorer: https://fcd.terra.dev/
    """

    symbol = 'LUNA'
    base_url = 'https://fcd.terra.dev/v1'
    rate_limit = 0.5
    coef = Decimal(1e-6)
    max_items_per_page = 100
    page_offset_step = 1

    supported_requests = {
        'get_balance': '/bank/{address}',
        'get_delegations': '/staking/{address}',
        'get_txs': '/txs?account={address}&limit={limit}&page={page}'
    }

    symbols = {
        'uluna': 'LUNA',
        'ukrw': 'KRT',
        'usdr': 'SDT',
        'uusd': 'UST',
        'umnt': 'MNT'
    }

    tx_kinds = {
        'bank/MsgSend': 'deposit',
        'staking/MsgDelegate': 'delegation',
        'distribution/MsgWithdrawDelegationReward': 'delegation_withdrawal'
    }

    def get_balance(self):
        balances = self.request('get_balance', address=self.address)
        if not balances:
            return None

        return_balances = []
        for bal in balances['balance']:
            return_balances.append({
                'symbol': self._get_symbol(bal['denom']),
                'amount': Decimal(bal['available']) * self.coef
            })

        # Add staked amount in LUNA
        for delegation in balances['delegations']:
            luna = next((b for b in return_balances if b['symbol'] == 'LUNA'),
                        {'symbol': 'LUNA', 'amount': 0})
            luna['amount'] += Decimal(delegation['amount']) * self.coef

        return return_balances

    def get_txs(self, offset=1, limit=100, unconfirmed=False):
        txs_req = self.request('get_txs', address=self.address, limit=limit,
                               page=offset)
        if not txs_req:
            return None

        txs = []
        for tx in txs_req['txs']:
            tx_item = {
                'kind': self.tx_kinds.get(tx['tx']['value']['msg'][0]['type']),
                'result': self.parse_tx(tx)
            }
            txs.append(tx_item)

        return txs

    def get_delegations(self):
        delegations = self.request('get_delegations', address=self.address)

        # Convert all numbers
        return self._load(delegations)

    def parse_tx(self, tx):
        fee = tx['tx']['value']['fee']
        msg = tx['tx']['value']['msg']
        return {
            'date': tx['timestamp'],
            'fee': [{
                'symbol': self._get_symbol(f['denom']),
                'amount': Decimal(f['amount']) * self.coef
            } for f in fee['amount']],
            'amount': [self.parse_tx_amount(m['value'])for m in msg]
        }

    def parse_tx_amount(self, tx_value):

        if 'amount' not in tx_value:
            return None

        tx_amount = tx_value['amount']
        if isinstance(tx_amount, list):
            amount = [{
                'symbol': self._get_symbol(t['denom']),
                'amount': Decimal(t['amount']) * self.coef
            } for t in tx_amount]
        else:
            amount = {
                'symbol': self._get_symbol(tx_amount['denom']),
                'amount': Decimal(tx_amount['amount']) * self.coef
            }
        return amount

    @classmethod
    def _get_symbol(cls, denom):
        """It seems that API returns only denom instead of correct
        symbols.
        """
        return cls.symbols.get(denom, 'unknown')



from blockapi.services import BlockchainAPI


class BlockscoutAPI(BlockchainAPI):
    """
    Multi coins: ethereum classic, xDAI, POA
    API docs: https://blockscout.com/{chain}/{network}/api_docs
    Explorer: https://blockscout.com
    """

    active = True

    rate_limit = 0
    coef = 1e-18
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'eth_get_balance': '?module=account&action=eth_get_balance'
                           '&address={address}',
        'get_balance': '?module=account&action=balance&address={address}',
        'get_token_list': '?module=account&action=tokenlist&address={address}',
        'get_token_balance': '?module=account&action=tokenbalance'
                             '&contractaddress={contract}&address={address}'
    }

    def eth_get_balance(self):
        response = self.request('eth_get_balance', address=self.address)
        if not response:
            return []

        amount = int(response['result'], 16) * self.coef
        return [{
            'symbol': self.symbol,
            'amount': amount
        }]

    def get_balance(self):
        balances = []

        main_balance = self._get_main_balance()
        if main_balance:
            balances.append(main_balance)

        token_balances = self._get_token_balances()
        if token_balances:
            balances += token_balances

        return balances

    def _get_main_balance(self):
        response = self.request('get_balance', address=self.address)
        if response['message'] != 'OK':
            return

        return {
            'symbol': self.symbol,
            'amount': int(response['result']) * self.coef
        }

    def _get_token_balances(self):
        response = self.request('get_token_list', address=self.address)
        if response['message'] != 'OK':
            return

        return [self._parse_token_balance(t)
                for t in response['result'] if int(t['balance'])]

    @staticmethod
    def _parse_token_balance(raw):
        decimals = int(raw.get('decimals', 0))

        return {
            'symbol': raw.get('symbol', 'UNKNOWN'),
            'address': raw['contractAddress'],
            'amount': (float(raw['balance']) * pow(10, -decimals)
                       if decimals else float(raw['amount'])),
            'name': raw.get('name', 'Unknown'),
            'type': raw.get('type', 'UNKNOWN'),
            'decimals': decimals
        }


class BlockscoutEthereumClassicAPI(BlockscoutAPI):
    """
    xDAI
    API docs: https://blockscout.com/etc/mainnet/api
    Explorer: https://blockscout.com/etc/mainnet
    """

    symbol = 'ETC'
    base_url = 'https://blockscout.com/etc/mainnet/api'


class BlockscoutXdaiAPI(BlockscoutAPI):
    """
    xDAI
    API docs: https://blockscout.com/poa/xdai/api
    Explorer: https://blockscout.com/poa/xdai
    """

    symbol = 'STAKE'
    base_url = 'https://blockscout.com/poa/xdai/api'



from decimal import Decimal

from pytest import mark

from blockapi.api.terra_money import TerraMoneyApi
from blockapi.test_data import test_addresses


class TestTerraMoneyApi:

    ADDRESS = test_addresses["LUNA"][0]

    def test_init(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        assert api

    @mark.vcr()
    def test_get_balance(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        result = api.get_balance()

        assert result == [
            {
                'symbol': 'LUNA',
                'amount': Decimal('34682.50182799999843055130568')
            },
            {
                'symbol': 'KRT',
                'amount': Decimal('9412.963716999999574045618491')
            },
            {
                'symbol': 'SDT',
                'amount': Decimal('0.05280299999999999761056454874')
            },
            {
                'symbol': 'UST',
                'amount': Decimal('0.01211199999999999945190913044')
            },
            {
                'symbol': 'MNT',
                'amount': Decimal('1122.501886999999949204670134')
            }
        ]

    @mark.vcr()
    def test_get_txs(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        result = api.get_txs()

        assert len(result) == 11

    @mark.vcr()
    def test_get_delegations(self):
        api = TerraMoneyApi(address=self.ADDRESS)
        result = api.get_delegations()

        assert result



from os import environ

test_addresses = {
    'BTC': [
        '35hK24tcLEWcgNA4JxpvbkNkoAcDGqQPsP',
        'xpub6CUGRUonZSQ4TWtTMmzXdrXDtypWKiKrhko4egpiMZbpiaQL2jkwSB1icqYh2cfDfVxdx4df189oLKnC5fSwqPfgyP3hooxujYzAu3fDVmz'
        # 'ypub6WjHjrJLKSg8oQw1E4LGvQDJ2uofgMfJKLnv5Ha4NPRW4rf7LPXffMJ8EReixY1mUCc33SsiDUodUCTvCktFeN7ZW28GVhBXhNnoKYUqXbP',
    ],
    'BCH': [
        '35hK24tcLEWcgNA4JxpvbkNkoAcDGqQPsP',
    ],
    'BSV': [
        '1M4CZGNmAv3o33wYoMLATJp4y4BTDe2okd'
    ],
    'ADA': [
        'DdzFFzCqrhtCYb5skdVUvyKJ7AohbUesTsTB4GQbfqRmXU4X7Y2v6YwQCjANWcF3Add7yVn3JqtAC2n2jPZymEX5k7w2gFFnsaWpfZkW'
    ],
    'EOS': ['fzrv13xeykgc'],
    'ETH': [
        '0xca8fa8f0b631ecdb18cda619c4fc9d197c8affca'
    ],
    'LTC': [
        'M8T1B2Z97gVdvmfkQcAtYbEepune1tzGua'
    ],
    'XTZ': [
        'tz1bDXD6nNSrebqmAnnKKwnX1QdePSMCj4MX',
        'KT1CRN88fn8EdWmZT9954Xf8NDsGfxtuwome'
    ],
    'DCR': [
        'DsjshD6PK3rgcKZtNUP7jnMcGQRvbY3dmwo',
        'DcrP3rA8RSi8Ai9VZU2pB2m31c9GWhEqii8'
        # current api doesn't support xpub key
        # 'dpubZFwoKEEJYVDxGo8bf2E4qwh6Qve9cku5gvaS5kC96hUdMT7SF9nymaLFeEFQaHy8a3SuiUJRL87rz3bfwFSFqErYVeHUg3xnzPjHftiofFu',
    ],
    'ATOM': [
        'cosmos1gn326f6sza44xt5kxrsdrnapp2sxhav03rhcsz'
    ],
    'NEO': [
        'AZnTM3mYbx9yzg8tb6hr7w9pAKntDmrtqk'
    ],
    'DOGE': [
        'DH5yaieqoZN36fDVciNyRueRGvGLR3mr7L'
    ],
    'ZEC': [
        't1VShHAhsQc5RVndQLyM1ZbQXLHKd35GkG1'
    ],
    'DASH': [
        'XtAG1982HcYJVibHxRZrBmdzL5YTzj4cA1'
    ],
    'ETC': [
        '0x618F37D7ff7B140E604172466CD42D1Ec35E0544'
    ],
    'ZEN': [
        'znZTLu1asaLWxB7EBqBRQ6DCnNyctYA3Rm4'
    ],
    'BNB': [
        'bnb1jxfh2g85q3v0tdq56fnevx6xcxtcnhtsmcu64m'
    ],
    'XLM': [
        'GDD7ABRF7BCK76W33RXDQG5Q3WXVSQYVLGEMXSOWRGZ6Z3G3M2EM2TCP'
    ],
    'RVN': [
        'RK6wMGc1tvLNDf4LoW6uoEcdUJVRfKsC7h'
    ],
    'TRX': [
        'TNNc1HGDUrRkowQxdcUaWyBodZXshuVtBp'
    ],
    'GRS': [
        'Fr5m2irs9vNWSAFXJK6KPtxqW9YWg384FX'
    ],
    'VET': [
        '0xcE6b1252b32a34fc4013F096cDf90643fB5D23ba'
    ],
    'BOS': [
        'GCPQQIX2LRX2J63C7AHWDXEMNGMZR2UI2PRN5TCSOVMEMF7BAUADMKH5'
    ],
    'LUNA': [
        'terra1vw96exejy03w7n2jt8t3qv392f3lqtvmv53720'
    ],
    'SOL': [
        '31dpiondDhZaqK23Re8kzkhY6CFEG9ZTQnr3shQm7g8b'
    ],
    'ONT': [
        'AS5R2tAn2KViJN3W4KkW5XwrrR85sXpn9p'
    ],
    'STAKE': [
        '0x6a1cf24C645DB2e37141Fa12E70Cb67e56b336f3'
    ],
    'DOT': [
        '13vrWSUXdm9yJH4GPig2NDMJFxUHwepeUbHgBr6Kmya9eVLh'
    ],
    'WND': [
        '5GeeixzC3xcCGc4Usi2UnCTfh46WXBmVL1duL9Nx5NHJkGxs'
    ]
}


def get_test_api_key(api_cls_name):
    """Tries to load api key from environment."""
    if api_cls_name.startswith('CryptoID'):
        key_field = 'CRYPTOIDAPI_KEY'
    else:
        key_field = f'{api_cls_name.upper()}_KEY'

    return environ.get(key_field)



import json
from datetime import datetime
from decimal import Decimal

from blockapi.services import BlockchainAPI, AddressNotExist, APIError
from blockapi.utils.decimal import safe_decimal


class SubscanAPI(BlockchainAPI):
    """
    API docs: https://docs.api.subscan.io/
    Explorer: https://www.subscan.io
    """

    api_url = 'api.subscan.io'
    rate_limit = 0
    coef = Decimal('1e-12')
    max_items_per_page = 50
    page_offset_step = 1
    confirmed_num = None
    xpub_support = True

    supported_requests = {
        'get_balance': '/api/v2/scan/search',
        'get_txs': '/api/scan/transfers',
        'get_rewards': '/api/scan/account/reward_slash',
        'get_staking': '/api/v2/scan/search'
    }

    def __init__(self, address, api_key=None, *args, **kwargs):
        super().__init__(address, api_key)
        self._headers = {
            'Content-Type': 'application/json'
        }

    def get_balance(self):
        body = json.dumps({
            'key': self.address,
            'row': 1,
            'page': 0,
        })

        response = self.request(
            'get_balance',
            body=body,
            headers=self._headers,
            api_key=self.api_key
        )

        if response['code'] == 0:
            return [{
                'symbol': self.symbol,
                'amount': safe_decimal(response['data']['account']['balance'])
            }]
        elif response['code'] == 10004:
            raise AddressNotExist()
        else:
            raise APIError(response['message'])

    def get_txs(self, offset=0, limit=10, unconfirmed=False):
        body = json.dumps({
            'address': self.address,
            'page': offset,
            'row': limit
        })

        response = self.request(
            'get_txs',
            body=body,
            headers=self._headers,
            api_key=self.api_key
        )

        if response['code'] == 0:
            transfers = (self._parse_tx(t) for t in response['data']['transfers'])
            return [t for t in transfers if unconfirmed or t['confirmed']]
        elif response['code'] == 10004:
            raise AddressNotExist()
        else:
            raise APIError(response['message'])

    def _parse_tx(self, tx):
        if tx['from'] == self.address:
            direction = 'outgoing'
        else:
            direction = 'incoming'

        return {
            'date': datetime.fromtimestamp(tx['block_timestamp']),
            'from_address': tx['from'],
            'to_address': tx['to'],
            'amount': safe_decimal(tx['amount']),
            'fee': safe_decimal(tx['fee']) * self.coef,
            'hash': tx['hash'],
            'confirmed': tx['success'],
            'is_error': not tx['success'],
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'raw': tx
        }

    def get_rewards(self, offset=0, limit=20):
        body = json.dumps({
            'address': self.address,
            'row': limit,
            'page': offset
        })

        response = self.request(
            'get_rewards',
            body=body,
            headers=self._headers,
            api_key=self.api_key
        )

        if response['code'] == 0:
            return [self._parse_reward(r) for r in
                    (response['data']['list'] if response['data']['list'] is not None else [])]
        elif response['code'] == 10004:
            raise AddressNotExist()
        else:
            raise APIError(response['message'])

    def _parse_reward(self, reward):
        if reward['event_id'] == 'Reward':
            sign = 1
        else:
            sign = -1

        return {
            'amount': safe_decimal(reward['amount']) * self.coef * sign,
            'hash': reward['extrinsic_hash'],
            'event_index': reward['event_index']
        }

    def get_staking(self, offset=0, limit=20):
        body = json.dumps({
            'key': self.address,
            'row': limit,
            'page': offset
        })

        response = self.request(
            'get_staking',
            body=body,
            headers=self._headers,
            api_key=self.api_key
        )

        if response['code'] == 0:
            return safe_decimal(response['data']['account']['balance_lock'])
        elif response['code'] == 10004:
            raise AddressNotExist()
        else:
            raise APIError(response['message'])


class SubscanPolkaAPI(SubscanAPI):
    chain_name = 'polkadot'
    symbol = 'DOT'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanKusamaAPI(SubscanAPI):
    chain_name = 'kusama'
    symbol = 'KSM'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanDarwiniaAPI(SubscanAPI):
    chain_name = 'darwinia'
    symbol = 'RING'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanDarwiniaCrabAPI(SubscanAPI):
    chain_name = 'crab'
    symbol = 'CRING'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanChainXAPI(SubscanAPI):
    chain_name = 'chainx'
    symbol = 'PCX'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanCloverAPI(SubscanAPI):
    chain_name = 'clover'
    symbol = 'CLV'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanEdgewareAPI(SubscanAPI):
    chain_name = 'edgeware'
    symbol = 'EDG'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanCentrifugeAPI(SubscanAPI):
    chain_name = 'centrifuge'
    symbol = 'CFG'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanPlasmAPI(SubscanAPI):
    chain_name = 'plasm'
    symbol = 'PLM'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanDockAPI(SubscanAPI):
    chain_name = 'dock'
    symbol = 'DCK'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanKulupuAPI(SubscanAPI):
    chain_name = 'kulupu'
    symbol = 'KLP'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanStafiAPI(SubscanAPI):
    chain_name = 'stafi'
    symbol = 'FIS'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanDataHighwayAPI(SubscanAPI):
    chain_name = 'datahighway'
    symbol = 'DHX'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanEquilibriumAPI(SubscanAPI):
    chain_name = 'equilibrium'
    symbol = 'EQ'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanRococoV1API(SubscanAPI):
    chain_name = 'rococo'
    symbol = 'ROC'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanWestendAPI(SubscanAPI):
    chain_name = 'westend'
    symbol = 'WND'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanMoonbaseAPI(SubscanAPI):
    chain_name = 'moonbase'
    symbol = 'DEV'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanCloverTestnetAPI(SubscanAPI):
    chain_name = 'clover-testnet'
    symbol = 'CLV'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanGatewayAPI(SubscanAPI):
    chain_name = 'gateway-testnet'
    symbol = 'CASH'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanAcalaMandalaAPI(SubscanAPI):
    chain_name = 'acala-testnet'
    symbol = 'ACA'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanPhalaVendettaAPI(SubscanAPI):
    chain_name = 'phala'
    symbol = 'tPHA'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanCrustMaxwellAPI(SubscanAPI):
    chain_name = 'crust'
    symbol = 'CRU'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanDataHighwayHarbourAPI(SubscanAPI):
    chain_name = 'datahighway-harbour'
    symbol = 'DHX'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'


class SubscanPangolinAPI(SubscanAPI):
    chain_name = 'pangolin'
    symbol = 'PRING'
    base_url = f'https://{chain_name}.{SubscanAPI.api_url}'



from pytest import mark

from blockapi.api.tzstats import TzStatsAPI
from blockapi.test_data import test_addresses


class TestTzscanAPI:
    ADDRESS = test_addresses["XTZ"][0]
    REWARD_ADDRESS = test_addresses["XTZ"][1]

    def test_init(self):
        api = TzStatsAPI(address=self.ADDRESS)
        assert api

    def test_get_balance(self):
        api = TzStatsAPI(address=self.ADDRESS)
        result = api.get_balance()

        assert result[0]['symbol'] == api.symbol
        assert result[0]['amount'] > 0

    @mark.vcr()
    def test_get_account(self):
        api = TzStatsAPI(address=self.REWARD_ADDRESS)
        result = api.get_account()

        assert "address" in result
        assert "manager" in result
        assert "delegate" in result



from blockapi.services import BlockchainAPI


class AmberdataAPI(BlockchainAPI):
    """
    Ethereum
    API docs: https://docs.amberdata.io/reference#get-all-addresses
    """

    symbol = 'ETH'
    base_url = 'https://web3api.io/api/v2'
    # 3 reqs per second with free api key, but doesn't work well
    rate_limit = 2
    coef = 1e-18
    start_offset = 0
    max_items_per_page = 100
    page_offset_step = 1

    supported_requests = {
        'get_balance': '/addresses/{address}/account-balances/latest',
        'get_token_balance': '/addresses/{address}/tokens?page={page}&size={size}',
        'get_token_info': '/addresses/{address}/information?include_price={include_price}'
    }

    def __init__(self, address, api_key):
        super().__init__(address, api_key)
        self._headers = {
            'x-amberdata-blockchain-id': 'ethereum-mainnet',
            'x-api-key': self.api_key
        }

    def get_balance(self):
        balances = []

        eth_balance = self._get_eth_balance()
        if eth_balance:
            balances.append(eth_balance)

        token_balances = self._get_token_balances()
        if token_balances:
            balances += token_balances

        return balances

    def _get_eth_balance(self):
        response = self.request(
            'get_balance',
            address=self.address,
            headers=self._headers
        )
        if response['status'] != 200:
            return

        return {
            'symbol': 'ETH',
            'amount': int(response['payload']['value']) * self.coef
        }

    def _get_token_balances(self):
        balances = []
        page = self.start_offset

        while True:
            response = self.request(
                'get_token_balance',
                address=self.address,
                page=page,
                size=self.max_items_per_page,
                headers=self._headers
            )
            if response['status'] != 200:
                break

            balances += [self._parse_token_balance(b)
                         for b in response['payload']['records']]

            if len(balances) == int(response['payload']['totalRecords']):
                break

            page += self.page_offset_step

        return balances

    @staticmethod
    def _parse_token_balance(raw):
        decimals = int(raw.get('decimals'))

        if raw['isERC20']:
            type_ = 'ERC-20'
        elif raw['isERC721']:
            type_ = 'ERC-721'
        elif raw['isERC777']:
            type_ = 'ERC-777'
        elif raw['isERC884']:
            type_ = 'ERC-884'
        elif raw['isERC998']:
            type_ = 'ERC-998'
        else:
            type_ = None

        return {
            'symbol': raw.get('symbol', 'UNKNOWN'),
            'address': raw['address'],
            'amount': (float(raw['amount']) * pow(10, -decimals)
                       if decimals else float(raw['amount'])),
            'name': raw.get('name', 'Unknown'),
            'type': type_
        }

    @classmethod
    def get_token_info(cls, token_address, api_key, include_price=False):
        api = cls(token_address, api_key)
        response = api.request(
            'get_token_info',
            address=token_address,
            include_price=include_price,
            api_key=api.api_key,
            headers=api._headers
        )
        if response['status'] != 200:
            return

        if response.get('payload'):
            return api._parse_token_info(token_address, response['payload'])

    @staticmethod
    def _parse_token_info(token_address, raw):
        symbol = raw['symbol'] if raw.get('symbol') else 'UNKNOWN'
        name = raw['name'] if raw.get('name') else 'Unknown'

        d = raw.get('decimals')
        decimals = int(d) if d else 0

        contract_types = raw.get('contractTypes', [])
        if 'Token' in contract_types or 'ERC20' in contract_types:
            type_ = 'ERC-20'
        elif 'ERC721' in contract_types:
            type_ = 'ERC-721'
        else:
            type_ = contract_types[0] if contract_types else 'Unknown'

        return {
            'symbol': symbol,
            'address': token_address,
            'name': name,
            'decimals': decimals,
            'type': type_
        }
